import asyncio
import os
import threading
import uuid
from matrix_wrapper import *
import json
from datetime import datetime, timedelta
import Timer

class Timer:
    CALLBACK_MESSAGE_JSON_ID = "callback_message"
    ID_JSON_ID = "id"
    END_JSON_ID = "end"
    TIME_JSON_FORMAT = "%d/%m/%Y/%H/%M/%S"

    STATE_CREATED = "created"
    STATE_RUNNING = "running"
    STATE_CANCELED = "canceled" # means stopped, a new timer has to be created for "restarting"
    STATE_FINISHED = "finished"

    timers_lock: threading.Lock = threading.Lock()
    timers: list[Timer] = []

    _timers_backup_file_lock: threading.Lock = threading.Lock()
    _backup_path: str = os.environ["TIMERS_PATH"]

    _id: str = None
    _timer: threading.Timer = None
    _callback_message: str = None
    _end_time: datetime = None
    _started_at: datetime = None
    _loop: asyncio.AbstractEventLoop = None
    _state: str = None
    _backup: bool = None
    _successfull_callbacks: list = None
    _cancel_callbacks: list = None

    def __init__(self, duration: float, callback_message: str=None, id: str=None, backup: bool=True, successfull_callbacks: list=[], cancel_callbacks: list=[]) -> None:
        self._callback_message = callback_message
        self._backup = backup
        self._successfull_callbacks = successfull_callbacks
        self._cancel_callbacks = cancel_callbacks
        self._timer = threading.Timer(duration, self._callback)
        self._loop = asyncio.get_event_loop()

        self._id = id if id is not None else str(uuid.uuid4())
        self._state = Timer.STATE_CREATED

        with Timer.timers_lock:
            Timer.timers.append(self)

    def _callback(self):
        if self.get_callback_message() is not None:
            asyncio.run_coroutine_threadsafe(send_message(self.get_callback_message()), self._loop)
        self._remove()
        self._state = Timer.STATE_FINISHED

        for callback in self._successfull_callbacks:
            asyncio.run_coroutine_threadsafe(callback(self), self._loop)

    def start(self):
        if self._timer.is_alive(): # already started
            return

        self._started_at = datetime.now()
        self._end_time = self._started_at + timedelta(seconds=self._timer.interval)
        self._timer.start()
        self._state = Timer.STATE_RUNNING

        if self._backup:
            self._dump()

    def cancel(self): # after cancel, a new timer has to be created to "restart"
        self._timer.cancel()
        self._remove()
        self._state = Timer.STATE_CANCELED

        for callback in self._cancel_callbacks:
            callback(self)

    def elapsed(self):
        return datetime.now() - self._started_at

    def remaining(self):
        return timedelta(seconds=self._timer.interval - self.elapsed().total_seconds())

    
    def _remove_backup(self):
        if not os.path.exists(Timer._backup_path):
            return

        with Timer._timers_backup_file_lock:
            timers = []
            with open(Timer._backup_path, "r") as f:
                for line in f:
                    timer = json.loads(line)
                    if timer.get(Timer.ID_JSON_ID) != self._id:
                        timers.append(timer)

            with open(Timer._backup_path, "w") as f:
                for timer in timers:
                    f.write(Timer.get_dump_str(timer))

    def _remove(self):
        self._remove_backup
        with Timer.timers_lock:
            Timer.timers.remove(self)

    def _dump(self):
        dump_str = Timer.get_dump_str(self.get_dump())
        with Timer._timers_backup_file_lock:
            with open(Timer._backup_path, "a+") as f:
                f.write(dump_str)

    def get_dump(self):
        return {Timer.ID_JSON_ID: self.get_id(),
        Timer.CALLBACK_MESSAGE_JSON_ID: self.get_callback_message(), 
        Timer.END_JSON_ID: self._end_time.strftime(Timer.TIME_JSON_FORMAT)}

    def get_id(self):
        return self._id

    def get_callback_message(self):
        return self._callback_message

    def get_state(self):
        return self._state

    def add_callbacks(self, successful_callbacks=[], cancel_callbacks=[]):
        self._successfull_callbacks.extend(successful_callbacks)
        self._cancel_callbacks.extend(cancel_callbacks)

    @staticmethod
    async def restartExisting():
        if not os.path.exists(Timer._backup_path):
            return
        timers = []
        with Timer._timers_backup_file_lock:
            with open(Timer._backup_path, "r") as f:
                for line in f:
                    timers.append(json.loads(line))

        with Timer._timers_backup_file_lock:
            with open(Timer._backup_path, "r+") as f:
                f.truncate(0)

        for timer in timers:
            created_timer = await Timer.check_and_create_timer(timer)
            if created_timer is not None:
                created_timer.start()

    @staticmethod 
    async def check_and_create_timer(timer_obj: dict, backup=True, successfull_callbacks: list=[], cancel_callbacks: list=[]) -> Timer:
        with Timer.timers_lock:
            for timer in Timer.timers:
                if timer.get_id() == timer_obj.get(Timer.ID_JSON_ID):
                    return timer
                

        timer_time = datetime.strptime(timer_obj.get(Timer.END_JSON_ID), Timer.TIME_JSON_FORMAT)
        current_time = datetime.now()

        if current_time >= timer_time:
            await send_message("timer has expired before it could be restarted: {}".format(timer_obj.get(Timer.CALLBACK_MESSAGE_JSON_ID)))
        else:
            remaining_time = timer_time - current_time
            return Timer(remaining_time.total_seconds(), timer_obj.get(Timer.CALLBACK_MESSAGE_JSON_ID), timer_obj.get(Timer.ID_JSON_ID), backup=backup, successfull_callbacks=successfull_callbacks, cancel_callbacks=cancel_callbacks) # each of them already backed up so you could assume another backup

        return None

    @staticmethod 
    def get_dump_str(serialized_object: dict) -> str:
        return json.dumps(serialized_object) + "\n"

    @staticmethod
    def get_from_id(timer_id: str):
        with Timer.timers_lock:
            for task in Timer.timers:
                if task.get_id() == timer_id:
                    return task

        return None

